package ratelimit

import "container/heap"

type queueItem interface {
	Expiry() int64
	Index() int
	SetIndex(int)
}

// A priorityQueue implements heap.Interface and holds queueItems.
type priorityQueue []queueItem

func (pq priorityQueue) Len() int { return len(pq) }

func (pq priorityQueue) Less(i, j int) bool {
	return pq[i].Expiry() < pq[j].Expiry()
}

func (pq priorityQueue) Swap(i, j int) {
	pq[i], pq[j] = pq[j], pq[i]
	pq[i].SetIndex(i)
	pq[j].SetIndex(j)
}

func (pq *priorityQueue) Push(x any) {
	n := len(*pq)
	item := x.(queueItem)
	item.SetIndex(n)
	*pq = append(*pq, item)
}

func (pq *priorityQueue) Pop() any {
	old := *pq
	n := len(old)
	item := old[n-1]
	old[n-1] = nil    // avoid memory leak
	item.SetIndex(-1) // for safety
	*pq = old[0 : n-1]
	return item
}

// update modifies the priority and value of an Item in the queue.
func (pq *priorityQueue) update(item queueItem) {
	heap.Fix(pq, item.Index())
}
